#include "common.h"
#include "SampleGenerator.h"
#include "Halton.h"

using namespace std;

//SampleGenerator::SampleGenerator() {}

SampleGenerator::SampleGenerator(int numSamples, float radius):
mNumSamples(numSamples), mRadius(radius)
{}


PoissonDiscSampleGenerator::PoissonDiscSampleGenerator(int numSamples, float radius):
SampleGenerator(numSamples, radius)
{}


void PoissonDiscSampleGenerator::Generate(Sample2 *samples) const
{
	// this is a hacky poisson sampling generator which does random dart-throwing
	// until it is not able to place any dart for a number of tries
	// in this case, the required min distance is reduced
	// the solution is a possion sampling with respect to the adjusted min distance
	// better solutions have been proposed, i.e., using hierarchical sampling

	const float maxTries = 1000;
	const float f_reduction = 0.9f;

	static HaltonSequence halton;
	float r[2];

	// generates poisson distribution on disc
	float minDist = 2.0f / sqrt((float)mNumSamples);

	//cout << "minDist before= " << minDist << endl;

	for (int i = 0; i < mNumSamples; ++ i)
	{
		int tries = 0, totalTries = 0;

		// repeat until valid sample was found
		while (1)
		{
			++ tries;
			++ totalTries;

			halton.GetNext(2, r);

			const float rx = r[0] * 2.0f - 1.0f;
			const float ry = r[1] * 2.0f - 1.0f;

			// check if in disk, else exit early
			if (rx * rx + ry * ry > 1)
				continue;

			bool sampleValid = true;

			// check poisson property
			for (int j = 0; ((j < i) && sampleValid); ++ j)
			{
				const float dist = 
					sqrt((samples[j].x - rx) * (samples[j].x - rx) +
					     (samples[j].y - ry) * (samples[j].y - ry));
			
				if (dist < minDist)
					sampleValid = false;
			}

			if (sampleValid)
			{
				samples[i].x = rx;
				samples[i].y = ry;
				break;
			}

			if (tries > maxTries)
			{
				minDist *= f_reduction;
				tries = 0;
			}
		}
		
		//cout << "sample: " << samples[i].x << " " << i << " " << samples[i].y << " r: " << sqrt(samples[i].x * samples[i].x  + samples[i].y * samples[i].y) << " tries: " << totalTries << endl;
	}

	//cout << "minDist after= " << minDist << endl;
}



GaussianSampleGenerator::GaussianSampleGenerator(int numSamples, float radius):
SampleGenerator(numSamples, radius)
{}


void GaussianSampleGenerator::Generate(Sample2 *samples) const
{
	static HaltonSequence halton;
	float r[2];

	const float sigma = mRadius;

	const float p0 = 1.0f / (sigma * sqrt(2.0f * M_PI));
	const float p1 = 1.0f / (sigma * sqrt(2.0f * M_PI));

	for (int i = 0; i < mNumSamples; ++ i)
	{
		halton.GetNext(2, r);

		float exp_x = -(r[0] * r[0]) / (2.0f * sigma * sigma);
		float exp_y = -(r[1] * r[1]) / (2.0f * sigma * sigma);

		samples[i].x = p0 * pow(M_E, exp_x);
		samples[i].y = p1 * pow(M_E, exp_y);
	}

	//cout << "minDist after= " << minDist << endl;
}


SphericalSampleGenerator::SphericalSampleGenerator(int numSamples, float radius):
SampleGenerator(numSamples, radius)
{}




void SphericalSampleGenerator::Generate(Sample2 *samples) const
{
	static HaltonSequence halton;
	float r[2];

	for (int i = 0; i < mNumSamples; ++ i)
	{
		halton.GetNext(2, r);

		// create stratified samples over sphere
		const float rx = r[0];
		const float ry = r[1];

		const float theta = 2.0f * acos(sqrt(1.0f - rx));
		const float phi = 2.0f * M_PI * ry;

		float x = sin(theta) * cos(phi);
		float y = sin(theta) * sin(phi);
		float z = cos(theta);

		// project to disc
		samples[i].x = x / z;
		samples[i].y = y / z;
	}
}