#include "ShadowMapping.h"
#include "FrameBufferObject.h"
#include "RenderState.h"
#include "RenderTraverser.h"
#include "Light.h"
#include <IL/il.h>
#include <assert.h>

using namespace std;


namespace CHCDemoEngine
{


static void PrintGLerror(char *msg)
{
	GLenum errCode;
	const GLubyte *errStr;
	
	if ((errCode = glGetError()) != GL_NO_ERROR) 
	{
		errStr = gluErrorString(errCode);
		fprintf(stderr,"OpenGL ERROR: %s: %s\n", errStr, msg);
	}
}



static CGprogram sCgShadowProgram;
static CGparameter sShadowParam;


static void GrabDepthBuffer(float *data, GLuint depthTexture)
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, depthTexture);

	glGetTexImage(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, GL_FLOAT, data);

	glBindTexture(GL_TEXTURE_2D, 0);
	glDisable(GL_TEXTURE_2D);
}


static void ExportDepthBuffer(float *data, int size)
{
	ilInit();
	assert(ilGetError() == IL_NO_ERROR);

	ILstring filename = ILstring("shadow.tga");
	ilRegisterType(IL_FLOAT);

	const int depth = 1;
	const int bpp = 1;

	if (!ilTexImage(size, size, depth, bpp, IL_LUMINANCE, IL_FLOAT, data))
	{
		cerr << "IL error " << ilGetError() << endl;
	
		ilShutDown();
		assert(ilGetError() == IL_NO_ERROR);

		return;
	}

	if (!ilSaveImage(filename))
	{
		cerr << "TGA write error " << ilGetError() << endl;
	}

	ilShutDown();
	assert(ilGetError() == IL_NO_ERROR);

	cout << "exported depth buffer" << endl;
}



ShadowMap::ShadowMap(Light *light, int size, const AxisAlignedBox3 &sceneBox, Camera *cam):
mSceneBox(sceneBox), mSize(size), mCamera(cam), mLight(light)
{
	mFbo = new FrameBufferObject(size, size, FrameBufferObject::DEPTH_32, true);
	// the diffuse color buffer
	mFbo->AddColorBuffer(ColorBufferObject::BUFFER_UBYTE, ColorBufferObject::WRAP_CLAMP_TO_EDGE, ColorBufferObject::FILTER_LINEAR, false);

	mShadowCam = new Camera(mSceneBox.Size().x * 0.5f, mSceneBox.Size().y * 0.5f);
	mShadowCam->SetOrtho(true);
}


ShadowMap::~ShadowMap()
{
	DEL_PTR(mFbo);
	DEL_PTR(mShadowCam);
}


void ShadowMap::ComputeShadowMap(RenderTraverser *renderer)
{
	const float xlen = Magnitude(mSceneBox.Diagonal() * 0.5f);
	const float ylen = Magnitude(mSceneBox.Diagonal() * 0.5f);
	
	mShadowCam->SetDirection(mLight->GetDirection());
	
	// set position so that we can see the whole scene
	Vector3 pos = mSceneBox.Center();

	pos -= mLight->GetDirection() * Magnitude(mSceneBox.Diagonal() * 0.5f);
	mShadowCam->SetPosition(pos);

	mFbo->Bind();
	
	glDrawBuffers(1, mrt);

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, mSize, mSize);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	glPolygonOffset(1.0f, 2000.0f);
	glEnable(GL_POLYGON_OFFSET_FILL);

	glShadeModel(GL_FLAT);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	// setup projection
	glOrtho(+xlen, -xlen, +ylen, -ylen, 0.0f, Magnitude(mSceneBox.Diagonal())); 
	

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// setup shadow camera
	mShadowCam->SetupCameraView();


	//////////////
	//-- compute texture matrix

	Matrix4x4 lightView, lightProj;

	mShadowCam->GetModelViewMatrix(lightView);
	mShadowCam->GetProjectionMatrix(lightProj);

	static Matrix4x4 biasMatrix(0.5f, 0.0f, 0.0f, 0.5f,
								0.0f, 0.5f, 0.0f, 0.5f,
								0.0f, 0.0f, 0.5f, 0.5f,
								0.0f, 0.0f, 0.0f, 1.0f); //bias from [-1, 1] to [0, 1]

	Matrix4x4 lightProjView = lightView * lightProj;
	mTextureMatrix = lightProjView * biasMatrix; 


	/////////////
	//-- render scene into shadow map

	renderer->RenderScene();

	
	glDisable(GL_POLYGON_OFFSET_FILL);
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();
	
	/*float *data = new float[mSize * mSize];

	GrabDepthBuffer(data, mFbo->GetDepthTex());
	ExportDepthBuffer(data, mSize);

	delete [] data;
	*/
	PrintGLerror("shadow map");

	FrameBufferObject::Release();
}


void ShadowMap::GetTextureMatrix(Matrix4x4 &m) const
{
	m = mTextureMatrix;
}

 
unsigned int ShadowMap::GetShadowTexture() const
{
	return mFbo->GetDepthTex();
}


} // namespace
